'''Classes to access Google search results.
Only the things we might actually be interested in seeing are exposed.
'''
from xml.dom import minidom
from Products.GoogleSATool.config import GoogleError
from DateTime import DateTime
from textwrap import TextWrapper
from Products.PythonScripts.standard import html_quote
from Products.CMFPlone.PloneBatch import Batch as PloneBatch
from XmlWrappers import XmlValueProperty, XmlAttributeProperty, XmlListProperty, XmlDictProperty, XmlWrapper

wrapper = TextWrapper(width=80)

__all__ = ['parse', 'parseString', 'GoogleResults',
    'DummyBatch', 'GoogleFS', 'GoogleGM', 'GoogleMT', 'GoogleNB', 'GooglePARAM',
    'GoogleR', 'GoogleRES', 'GoogleSpelling', 'GoogleOneSynonym', 'GoogleSynonyms']

class DummyBatch(PloneBatch):
    """A batch, but sequence elements are not accessible"""
    def __init__(self, size, first, last, length, next=None, prev=None):
        self._next = next
        self._prev = prev
        PloneBatch.__init__(self, SparseList([], first, length), size,
            first, last, orphan=0, overlap=0, quantumleap=0,
            b_start_str='start')
        
    def next(self):
        if not self._next:
            return None

        length = self.sequence_length
        return DummyBatch(self.size, self.end+1, min(length, self.end+self.size+1), length)

    def previous(self):
        if not self._prev:
            return None

        if self.first != 0:
            first = max(0, self.first - self.size)
            return DummyBatch(self.size, first, self.first, self.sequence_length)
        return None


class SparseList(object):
    """Like a list, but only a little bit in the middle actually exists"""
    def __init__(self, sequence, start, length):
        self.sequence = sequence
        self.start = start
        self.length = length

    def __len__(self):
        return self.length
    def __getitem__(self, index):
        if index < 0 or index >= self.length:
            raise IndexError()
        return None

class GoogleBLOB(XmlWrapper):
    def __str__(self):
        text = XmlWrapper.__str__(self)
        return text.encode('ascii').decode('base64')

    encoding = XmlAttributeProperty('encoding', '''The encoding scheme of the HTML data''')

class GoogleContainsBlob(XmlWrapper):
    '''A tag which contains a BLOB'''
    BLOB = XmlValueProperty('BLOB', klass=GoogleBLOB)

class GoogleC(XmlWrapper):
    SZ = XmlAttributeProperty('SZ', '''Provides the size of the cached version of the search result in kilobytes ("k").''')

    CID = XmlAttributeProperty('CID', '''Identifier of a document in
        Google's cache. To fetch the document from the cache, send a search
        term built like this: "cache:" + CID text + ":" + escaped URL. The
        escaped URL is available in the UE tag. Send this search term
        normally, as one would type it into the search form''')

class GoogleCACHE_LEGEND_TEXT(GoogleContainsBlob):
    fgcolor = XmlAttributeProperty('fgcolor', '''The foreground color of the query term.''')
    bgcolor = XmlAttributeProperty('bgcolor', '''The background colour of the query term.''')

class GoogleCACHE_LEGEND_FOUND(XmlWrapper):
    '''Provides encapsulation for query terms found in the visible text of the cached result returned'''
    CACHE_LEGEND_TEXT = XmlListProperty('CACHE_LEGEND_TEXT', '''Details of a query term which is visible in the cached result. Any query terms found in the cached result will automatically be highlighted using the colors described in the attributes of this tag.''',
        klass=GoogleCACHE_LEGEND_TEXT)
    
class GoogleCACHE(XmlWrapper):
    '''Provides encapsulation for the cached version of a search result'''
    CACHE_URL = XmlValueProperty('CACHE_URL', 'Initial URL of cached result')
    CACHE_REDIR_URL = XmlValueProperty('CACHE_REDIR_URL', 'Final URL of cached result after all redirects are resolved')
    CACHE_LAST_MODIFIED = XmlValueProperty('CACHE_LAST_MODIFIED',
        'Date that the document was crawled')
    CACHE_LEGEND_FOUND = XmlValueProperty('CACHE_LEGEND_FOUND', klass=GoogleCACHE_LEGEND_FOUND)
    CACHE_LEGEND_NOTFOUND = XmlValueProperty('CACHE_LEGEND_NOTFOUND', klass=GoogleContainsBlob)
    CACHE_CONTENT_TYPE = XmlValueProperty('CACHE_CONTENT_TYPE', '''MIME type of the cached result''')
    CACHE_LANGUAGE = XmlValueProperty('CACHE_LANGUAGE', '''The language of the cached result as determined by Google's automatic language classification algorithm.''')
    CACHE_ENCODING = XmlValueProperty('CACHE_ENCODING', '''The encoding scheme of the cached result''')
    CACHE_HTML = XmlValueProperty('CACHE_HTML', klass=GoogleContainsBlob)

class GoogleFS(XmlWrapper):
    '''Additional search result details'''
    NAME = XmlAttributeProperty('NAME', 'Name of the result descriptor')
    VALUE = XmlAttributeProperty('VALUE', 'Value of the result descriptor')

class GoogleGM(XmlWrapper):
    """Provides encapsulation for a single KeyMatch result"""
    # GL, GD?
    GL = XmlValueProperty('GL', '''Contains the URL of a KeyMatch result''')
    GD = XmlValueProperty('GD', '''Contains the description of a KeyMatch result''')

class GoogleHAS(XmlWrapper):
    L = XmlValueProperty('L', '''Indicates that the "link:" special query term is supported for this search result URL''',
        default=False, klass=lambda s:True)
    C = XmlValueProperty('C',
        '''Indicates that the "cache:" special query term is supported for this search result URL''',
        klass=GoogleC)

class GoogleMT(XmlWrapper):
    '''Meta tag name and value pairs pulled from the search result'''
    N = XmlAttributeProperty('N', '''Name of the meta tag''')
    V = XmlAttributeProperty('V', '''Value of the meta tag''')

class GoogleNB(XmlWrapper):
    PU = XmlValueProperty('PU', '''Contains relative URL to the previous results page''', default=None)
    NU = XmlValueProperty('NU', '''Contains relative URL to the next results page''', default=None)

class GooglePARAM(XmlWrapper):
    '''The input parameters submitted to the Google search engine to generate these results'''
    # Attributes
    name = XmlAttributeProperty('name',
        '''Input parameter name''')
    value = XmlAttributeProperty('value',
        '''HTML formatted version of the input parameter value''')
    original_value = XmlAttributeProperty('original_value',
        '''Original URL-escaped version of the input parameter value''')

    def sortkey(self):
        '''Parse sort key'''
        key = self.value.split(':') + ['']*4
        # key[0] 'date' is only valid value: ignore
        # key[1] 'A' or 'D'
        direction, mode, format = 'A', 'L', 'd1'
        
        if key[1] in ('A', 'D'):
            direction = key[1]
        if key[2] in ('S','R','L'):
            mode = key[2]

        return dict(direction=direction, mode=mode, format=format, relevance=mode=='L')

class GoogleR(XmlWrapper):
    '''Provides encapsulation for the details of an individual search result'''

    # Attributes
    N = XmlAttributeProperty('N', '''Indicates the index (1-based) of this search result''', int)
    L = XmlAttributeProperty('L', '''Indicates the recommended indentation level of the results.''', int, 1)
    # Note: Currently this value will always be 1 unless directory
    # crowding occurs. In this case, the second directory result will
    # have a value of 2.
    MIME = XmlAttributeProperty('MIME', '''Indicates the MIME type of the search result''')

    # Elements
    # U, T?, RK, FS?, MT*, S?, HAS, HN?
    U = XmlValueProperty('U', '''The URL of the search result.''')
    UE = XmlValueProperty('UE', '''undocumented: escaped URL?''')
    T = XmlValueProperty('T', '''The title of the search result''', default='')
    RK = XmlValueProperty('RK', '''Provides a general rating of the relevance of the search result''', int)
    FS = XmlDictProperty('FS', 'NAME', klass=GoogleFS)
    MT = XmlDictProperty('MT', 'N', klass=GoogleMT)
    S = XmlValueProperty('S', '''Search result snippet for the search result
        Note: Query terms will be in highlighted in bold in the
        results, and line breaks will be included for proper text wrapping.''')

    HAS = XmlValueProperty('HAS',
        '''Provides encapsulation for any special features supported for this search request''',
        klass=GoogleHAS)
    
    HN = XmlValueProperty('HN', '''Indicates that directory crowding
        has occurred and that additional results are available from the
        directory where this search result was found. The value of this tag is
        ready to be used with the "site:" query term.''')
    def wrapUrl(self):
        '''Return the URL with wordbreaks, and without the protocol'''
        url = self.strippedUrl().replace('/', '/ ')
        wrapped = '\n'.join(wrapper.wrap(url)).replace('/ ', '/')
        return wrapped
    def strippedUrl(self):
        '''Remove the protocol from the URL'''
        return self.U.split('://',1)[-1]

class GoogleRES(XmlWrapper):
    '''Provides encapsulation for the details of the individual search results'''

    # Attributes
    SN = XmlAttributeProperty('SN',
        '''Indicates the index (1-based) of the first search result returned in this result set''', int)
    EN = XmlAttributeProperty('EN',
        '''Indicates the index (1-based) of the last search result returned in this result set''', int)

    # Elements
    # M, FI?, XT?, NB?, R*
    M = XmlValueProperty('M', '''The estimated total number of results for the search
        Note: The estimate of the total number of results for a search can be too high or too low.''', int)
    FI = XmlValueProperty('FI',
        '''Indicates that document filtering was performed during this search''',
        klass=lambda n:True,
        default=False)
    NB = XmlValueProperty('NB', 'Batch navigation', klass=GoogleNB, default=None)
    R = XmlListProperty('R', klass=GoogleR)

    def Batch(self, num=10, limit=1000):
        '''Return a dummy batch object for navigation'''
        SN, EN = self.SN, self.EN
        NB = self.NB
        if NB is not None:
            next, prev = NB.NU, NB.PU
        else:
            next, prev = None, None
        length = min(self.M, limit-1)
        count = self.EN-self.SN+1
        # Reduce limit is there is no next batch: Google returns
        # unfiltered count even when we are filtering.
        if next is None or count < num:
            length = self.EN
        batch = DummyBatch(num, SN-1, EN-1, length, next, prev)
        return batch

class GoogleSpelling(XmlWrapper):
    '''Provides encapsulation for alternate spelling suggestions for the submitted query.
    Only one spelling suggestion is returned at this time.''',
    # Suggestion+
    class GoogleSuggestion(XmlWrapper):
        q = XmlAttributeProperty('q', 'The spelling suggestion quoted for URL')
        # string value is the HTML version.
        
    Suggestion = XmlListProperty('Suggestion', '', klass=GoogleSuggestion)

class GoogleOneSynonym(XmlWrapper):
    '''A single synonym suggestion'''
    q = XmlAttributeProperty('q', 'The synonym suggestion')

class GoogleSynonyms(XmlWrapper):
    '''Provides encapsulation for synonym suggestions for the
    submitted query. Up to 20 synonym suggestions may be returned
    depending on the synonym list associated with the front end by the
    search administrator.'''

    OneSynonym = XmlListProperty('OneSynonym',
        '''A synonym suggestion for the submitted query in HTML format.''',
        klass=GoogleOneSynonym)

class GoogleResults(XmlWrapper):
    '''Encapsulate GSA result XML in a Python Class'''
    # Attributes
    VER = XmlAttributeProperty('VER', '''Version''')

    # Elements
    # (TM, Q, PARAM*, CUSTOM?, Spelling?, Synonyms?, CT?, TT?, GM*, RES?) | CACHE
    TM = XmlValueProperty('TM', 'Total server time to return search results, measured in seconds.', float)
    Q = XmlValueProperty('Q', 'The search query submitted to the Google search engine to generate these results')
    PARAM = XmlDictProperty('PARAM', 'name', GooglePARAM)
    # CUSTOM: ignored
    Spelling = XmlValueProperty('Spelling', klass=GoogleSpelling)
    Synonyms = XmlValueProperty('Synonyms', klass=GoogleSynonyms)
    CT = XmlValueProperty('CT', '''Search comments e.g. Sorry, no content found for this URL''', default='')
    # TT? - not documented
    GM = XmlListProperty('GM', klass=GoogleGM)
    RES = XmlValueProperty('RES', klass=GoogleRES)

    # This could appear as an alternative to all of the above:
    CACHE = XmlValueProperty('CACHE', klass=GoogleCACHE)

    def __init__(self, dom):
        XmlWrapper.__init__(self, dom.getElementsByTagName('GSP')[0])
        assert self.VER=='3.2'
        #assert self._node.getAttribute('VER')=='3.2'

def parse(resultFile):
    dom = minidom.parse(resultFile)
    return GoogleResults(dom)

def parseString(result):
    dom = minidom.parseString(result)
    return GoogleResults(dom)
