# -*- coding: latin-1 -*-
import os, sys, re

# Zope
from Acquisition import aq_parent, aq_inner, aq_base
from OFS import SimpleItem
from OFS.PropertyManager import PropertyManager
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Products.CMFCore.FSPythonScript import FSPythonScript
from AccessControl import ClassSecurityInfo
from Globals import InitializeClass, package_home
from Products.CMFCore.utils import getToolByName
from Products.CMFCore.utils import UniqueObject
from Products.CMFCore.ActionProviderBase import ActionProviderBase
from Products.CMFCore.Expression import Expression, createExprContext
from Products.PageTemplates.Expressions import getEngine, SecureModuleImporter
from Products.PythonScripts.standard import html_quote
from ZTUtils import make_query
from zExceptions import Unauthorized

import urllib2, urllib
from DateTime import DateTime

from Products.GoogleSATool.config import METATYPE, TOOLNAME, TOOLTITLE, GoogleError
from Products.GoogleSATool.utils import *
from Products.GoogleSATool.GoogleSearch import DEFAULT_FIELDS, FieldDef, FieldValidation
from Products.GoogleSATool.GoogleResults import parse, parseString, GoogleResults
from Products.GoogleSATool.gpydoc import MyHTMLDoc

# Zope 3 interfaces, but on older zopes we'll just skip this bit.
try:
    from zope.interface import implements
except ImportError:
    def implements(*args): pass

from interfaces import IGoogleSATool, IGoogleSAToolManagement

__docformat__ = 'reStructuredText'
__all__ = ['GoogleSATool']

IGNORED_FORM_VARIABLES = dict.fromkeys(['submit', 'submit.x', 'submit.y', 'start'])

ERROR_TEMPLATE = '''<GSP VER="3.2">
<TM>0.210393</TM>
<Q>water</Q>
<CT>%s</CT>
</GSP>
'''

class GoogleSATool(UniqueObject, PropertyManager, 
                       SimpleItem.SimpleItem, ActionProviderBase,
                       FieldValidation):
    "The GoogleSATool tool"

    meta_type = METATYPE
    id = TOOLNAME
    title = TOOLTITLE
    implements(IGoogleSATool,IGoogleSAToolManagement)

    manage_options = PropertyManager.manage_options + \
                     SimpleItem.SimpleItem.manage_options + (
        {'label': 'Settings', 'action': 'google_settings',},
        {'label': 'Documentation', 'action': 'google_documentation',},
    )

    _properties = (
        {'id':'title', 'type':'string', 'mode':'w'},
    )

    # Standard security settings
    security = ClassSecurityInfo()

    security.declareProtected('Manage properties', 'google_settings')
    google_settings = PageTemplateFile('zpt/manage_google_settings', globals())
    google_settings.title = "%s settings" % TOOLTITLE

    security.declarePublic('docs')
    def docs(self):
        """Returns GSA docs formatted as HTML"""
        docpath = os.path.join(package_home(globals()), 'doc', 'documentation.txt')
        f = open(docpath, 'r')
        _docs = f.read()
        return _docs

    def moduledoc(self, modname=None):
        """Returns api documentation"""
        if modname is None:
            modname = 'Products.GoogleSATool.GoogleResults'

        if modname.endswith('.html'):
            modname = modname[:-5]
        if not modname.startswith('Products.GoogleSATool'):
            return self.google_moduledoc(
                content='Sorry, access to that class is not permitted.',
                base=self.REQUEST.ACTUAL_URL)
        module = None
        try:
            module = SecureModuleImporter[modname]
        except Unauthorized:
            content = 'Sorry, not a public interface'

        if module is not None:
            content = MyHTMLDoc().document(module)

        return self.google_moduledoc(content=content, base=self.REQUEST.ACTUAL_URL)

    security.declareProtected('Manage properties', "google_documentation")
    google_documentation = PageTemplateFile("zpt/google_documentation.pt", globals())
    google_documentation.title = 'Plone GSA documentation'

    security.declareProtected('Manage properties', "google_documentation")
    google_moduledoc = PageTemplateFile("zpt/google_moduledoc.pt", globals())
    google_moduledoc.title = 'Plone GSA API'

    security.declareProtected('Manage properties', "api")
    api = FSPythonScript('api',
        os.path.join(package_home(globals()), 'zpt', 'interface.py'))
    api.title = "Plone GSA API"
    
    def _flattenlist(self, value, separator):
        """converts value to a list:
           If already a list all inner values are flattened.
        """
        if isinstance(value, basestring):
            value = [value]
        res = []
        while value:
            v = value.pop()
            if isinstance(v, basestring):
                res.append(v)
            else:
                value = value + v
        return [v for v in reversed(res) if v]

    def _encodelist(self, value, separator):
        """URL encode a list for Google.
        The values need an additional encoding step in case they contain the separator.
        """
        escape = '%%%02X' % ord(separator)
        def encode(s):
            '''URL encoding including escaping the separator whatever it is.'''
            if isinstance(s, tuple):
                return ':'.join([urllib.quote_plus(part).replace(separator, escape)
                    for part in s])
            
            return urllib.quote_plus(s).replace('%3A',':', 1).replace(separator, escape)

        return separator.join([encode(v) for v in value ])

    security.declareProtected('Manage properties', 'manage_setGooglePrefs')
    def manage_setGooglePrefs(self, REQUEST=None,
        googleURL=None, fields=None, collectionRegex=None, searches=None):
        """
        Set the Google URL
        """
        if fields is not None:
            opts = self.getFieldOptions()
            for f in fields:
                if f.name:
                    opts[f.name] = FieldDef(f.name, f.default, getattr(f, 'required', False))
            self.compileFields(opts)
            self._fieldOptions = opts

        if googleURL is not None:
            self._googleURL = googleURL

        if collectionRegex is not None:
            self._collectionRegex = collectionRegex

        if searches is not None:
            self._extraSearches = tuple([
                (s.title, s.query) for s in searches if s.title])

        if REQUEST:
            #REQUEST.RESPONSE.redirect(self.absolute_url() + '/google_settings')
            REQUEST.portal_status_message = 'Settings saved at ' + str(DateTime())
            return self.google_settings(REQUEST)

    security.declareProtected('Manage properties', 'manage_deleteFields')
    def manage_deleteFields(self, REQUEST=None, fields=None, searches=None):
        """Delete the selected fields (provided they are deletable)"""
        if fields is not None:
            opts = self.getFieldOptions()
            todel = [ f.name for f in fields if f.delete and opts[f.name]['deletable'] ]
            for name in todel:
                del opts[name]
            self._fieldOptions = opts
        if searches is not None:
            self._extraSearches = tuple([
                (s.title, s.query) for s in searches
                if s.title and not s.delete])
            
        if REQUEST:
            REQUEST.portal_status_message = 'Deleted fields %s' % ', '.join(todel)
            return self.google_settings(REQUEST)
        
    security.declareProtected('View','getGoogleURL')
    def getGoogleURL(self):
        """
        Gets the Google URL
        """

        return getattr(self, '_googleURL', '')

    security.declarePrivate('compileFields')
    def compileFields(self, fields):
        '''Compile all the 'default value' expressions.'''
        for k in fields:
            field = fields[k]
            default = field['default']
            if default is not None and field.get('expr_default', None) is None:
                field['expr_default'] = Expression(default)

    def _getExpressionContext(self, object):
        portal = aq_parent(aq_inner(self))
        if object is None or not hasattr(object, 'aq_base'):
            folder = portal
        else:
            folder = object
            # Search up the containment hierarchy until we find an
            # object that claims it's a folder.
            while folder is not None:
                if getattr(aq_base(folder), 'isPrincipiaFolderish', 0):
                    # found it.
                    break
                else:
                    folder = aq_parent(aq_inner(folder))
        ec = createExprContext(folder, portal, object)
        return ec

    security.declareProtected('View', 'getFieldOptions')
    def getFieldOptions(self):
        """Gets the field settings"""
        options = getattr(self, '_fieldOptions', None)
        if options is None:
            options = DEFAULT_FIELDS
            self.compileFields(options)
            self._fieldOptions = options

        return options

    def getEditFieldOptions(self):
        '''Field options for edit screen, sorted and with a trailing blank entry.'''
        opts = sorted(self.getFieldOptions().itervalues(), key=(lambda v: v['name']))
        return opts + [FieldDef('')]

    security.declareProtected('View', 'getCollectionRegex')
    def getCollectionRegex(self):
        """The regular expression used to validate collection names"""
        return getattr(self, '_collectionRegex', '')
        
    security.declareProtected('View', 'getExtraSearches')
    def getExtraSearches(self):
        """Returns the canned searches as a list of dictionaries
        with keys:
        title: The canned search title.
        query: The query parameters from the configuration screen.
        decoded: A dictionary of parameters to be set or cleared in the query.
        url: The constructed URL.
        current: True if the current request could have generated this query.
        """
        REQUEST = self.REQUEST
        
        def decode(qs):
            bits = [ b.strip() for b in qs.split('&') if b.strip() and '=' in b ]
            res = {}
            for b in bits:
                name, value = b.split('=')
                value = urllib.unquote_plus(value)
                if value=='':
                    value = None
                res[name] = value
            return res

        def _isCurrentExtra(values):
            for name in values:
                formvalue = REQUEST.form.get(name, '')
                if values[name] is None and (formvalue is None or formvalue==''):
                    continue

                if formvalue != values[name]:
                    return False
            else:
                return True

        extra = []
        for title, query in getattr(self, '_extraSearches', ()):
            decoded = decode(query)
            extra.append(dict(title=title, query=query, decoded=decoded))

        # Need to exclude from the modified search any parameters
        # used in other modified searches.
        exclude = {}
        for e in extra:
            exclude.update(e['decoded'])
        for e in extra:
            d = e['decoded']
            for k in exclude:
                if k not in d:
                    d[k] = None
            e['url'] = self.make_query(**d)

        # Figure out which searches match the current search string.
        for e in extra:
            decoded = e['decoded']
            e['current'] = _isCurrentExtra(decoded)

        return extra

    security.declareProtected('View', 'getEditExtraSearches')
    def getEditExtraSearches(self):
        dummy = [dict(title='', query='')]
        return [ dict(title=title, query=query)
            for title, query in getattr(self, '_extraSearches', ())] + dummy

    security.declareProtected('View', 'matchExtraSearch')
    def matchExtraSearch(self):
        """Returns the title of the current canned search, or None.
        Can be used to say 'you searched in <some collection>'
        """
        REQUEST = self.REQUEST
        searches = self.getExtraSearches()
        for s in self.getExtraSearches():
            if s['current']:
                return s['title']
        return None

    security.declareProtected('View', 'search')
    def search(self, context, **kwargs):
        """
        Send a request to the Google appliance.
        'context' is passed through to Python expressions when building
        the query string.
        Values passed in kwargs take precedence over values calculated
        from the tool configuration.
        
        Returns the Google result XML mapped to a set of Python classes
        i.e. a GoogleResults object.
        """
        try:
            return self._search(context, **kwargs)
        except Exception, msg:
            return parseString(ERROR_TEMPLATE % html_quote(msg))

    def _search(self, context, **kwargs):
        searchURL = self.google_query(context, **kwargs)
#         print "searchURL", searchURL
        try:
            resultFile = urllib2.urlopen(searchURL)
#             print resultFile.read()
#             resultFile = urllib2.urlopen(searchURL)
        except IOError, msg:
            raise IOError('Google Server returned %s' % msg)

        try:
            results = parse(resultFile)
        finally:
            resultFile.close()

        #import pdb; pdb.set_trace()
        return results

    def make_query(self, url=None, **keywords):
        """Make a query string for Plone.
        Similar to ZTUtils make_query, but allows keyword=None to
        force an argument to be omitted.

        Also omits the 'submit' parameters automatically.
        """
        url = url or self.REQUEST.get('URL0')
        d = {}
        # We want to allow keyword=None to force an argument to be
        # omitted
        d.update(self.REQUEST.form)
        d.update(IGNORED_FORM_VARIABLES)
        d.update(keywords)
        
        # Now remove the bad ones...
        for k in d.keys():
            if d[k] is None:
                del d[k]

        mq = make_query(d)
        if mq:
            return '%s?%s' % (url, make_query(d))
        return url

    security.declareProtected('View', 'google_query')
    def google_query(self, context, **kwargs):
        """Make a Google query string from:
            Keyword arguments (highest priority)
            REQUEST form arguments (middle)
            configured defaults (lowest)
            
            Throws GoogleError if any required arguments are missing,
            or if any keyword arguments are not configured fields.
        """
        d = {}
        REQUEST = self.REQUEST
        form = REQUEST.form
        baseURL = self.getGoogleURL()
        expr_context = self._getExpressionContext(context)

        if not baseURL:
            raise GoogleError('Google URL not set')

        opts = self.getFieldOptions()
        for opt in opts.itervalues():
            name = opt['name']
            required = opt['required']
            expr = opt.get('expr_default', None)
            listsep = opt['listsep']
            value = None

            if name in kwargs:
                value = kwargs.pop(name)
                if listsep:
                    value = self._flattenlist(value, listsep)

            elif expr is not None:
                try:
                    value = expr(expr_context)
                except (KeyError, AttributeError):
                    #import traceback; traceback.print_exc()
                    value = None

                if listsep:
                    value = self._flattenlist(value, listsep)

                if value=='':
                    value = None
                if value is not None:
                    validator = getattr(self, 'validate_'+name, None)
                    if validator and not validator(value):
                        raise GoogleError('Invalid parameter value %r' % (value))

            if value is not None and value != '' and value != []:
                if listsep:
                    value = self._encodelist(value, listsep)
                if isinstance(value, bool):
                    value = int(value)
                d[name] = str(value)
            elif required:
                raise GoogleError('Missing required parameter %r' % name)

        if len(kwargs) != 0:
            raise GoogleError('invalid parameter(s) %r' % ', '.join(kwargs.keys()))

        query = '&'.join(['%s=%s' % (k, urllib.quote_plus(d[k])) for k in d])
        if '?' in baseURL:
            delim = '&'
        else:
            delim = '?'

        return "%s%s%s" % (baseURL, delim, query)

    security.declareProtected('View', 'getResultIcon')
    def getResultIcon(self, result):
        """Get an icon suitable for this result"""
        TEMPLATE = '<img class="gsa_icon" src="%s" alt="%s" title="%s" />'
        mime = result.MIME
        if mime:
            mtr = self.mimetypes_registry
            mtiList = mtr.lookup(mime)
            if len(mtiList) != 0:
                mti = mtiList[0]
                icon = mti.icon_path
                name = mti.name()
            return TEMPLATE % (icon, name, name)
        else:
            ptype = result.MT.get('DC.type', None)
            if ptype is not None:
                ptype = ptype.V;
                type_tool = getToolByName(self, 'portal_types')
                for t in type_tool.listTypeInfo():
                    if t.Title()==ptype:
                        return TEMPLATE % (t.getIcon(), ptype, ptype)

# Previewing will need support from content types. Probably best to
# add a 'preview' metadata field to the indexed content with the
# preview URL.
#
#     security.declareProtected('View', 'getPreview')
#     def getPreview(self, result):
#         ptype = result.MT.get('DC.type', None)
#         if ptype is None:
#             return
#         ptype = ptype.V
#         print ptype
#         if ptype=='Article':
#             return '/getThumbnail'


    security.declareProtected('View', 'encode')
    def encode(self, s):
        """Results are returned in unicode, this method will
        encode them to ascii converting other characters to
        character references."""
        return s.encode('ascii','xmlcharrefreplace')

    security.declarePublic("query_string")
    def query_string(self, replace={}, original=None):
        """ Updates 'original' dictionary by the values in the 'replace'
            dictionary and returns the result as url quoted query string.

            The 'original' dictionary defaults to 'REQUEST.form' if no
            parameter is passed to it. Keys in the 'replace' dictionary
            with a value of 'None' (or _.None in DTML) will be deleted
            from 'original' dictionary before being quoted.

            The original 'REQUEST.form' will remain unchanged.
        """
        # Based on code by Grgoire Weber
        if original is None:
            query = self.REQUEST.form.copy()
        else:
            query = original.copy()

        # delete key/value pairs if value is None
        for k,v in replace.items():
            if v is None:
                if query.has_key(k):
                    del query[k]
                del replace[k]

        # update dictionary
        query.update(replace)
        qs = '&'.join(["%s=%s" % (urllib.quote_plus(str(k)), urllib.quote_plus(str(v)))
            for k,v in query.items()])

        return qs

    security.declareProtected("View", "url_plus_query")
    def url_plus_query(self, url, query=None):
        """Adds query segment to an existing URL.
        Existing query parameters are may be overridden by query,
        otherwise they are preserved.
        """
        if query is None:
            query = {}
        parts = url.split('?', 1)
        oldargs = {}
        if len(parts) > 1:
            for arg in parts[1].split('&'):
                if '=' in arg:
                    k,v = [urllib.unquote_plus(s) for s in arg.split('=',1)]
                    oldargs[k] = v

        return "%s?%s" % (parts[0], self.query_string(query, oldargs))

InitializeClass(GoogleSATool)

