'''XML Wrappers
This file defines a bunch of property classes which can be used to
create simple Python classes wrapping an XML element.

See GoogleResults.py for an example of use.
'''

__all__ = ['XmlValueProperty', 'XmlAttributeProperty', 'XmlListProperty', 'XmlDictProperty', 'XmlWrapper']
_marker = object()

def docstring(doc, klass, modifier=''):
    """Add some type information to the docstring"""
    if doc is _marker:
        doc = klass.__doc__
    if klass is not None and isinstance(doc, basestring):
        doc += '\nReturns: '+modifier+klass.__name__
    return doc

def XmlValueProperty(tag, doc=_marker, klass=unicode, default=_marker):
    '''A property which returns the value of an XML node cast to a Python type'''
    if type(klass) is type:
        isValue = not issubclass(klass, XmlWrapper)
    else:
        isValue = True

    def accessor(self):
        nodes = self._node.getElementsByTagName(tag)
        if len(nodes)==0:
            if default is not _marker:
                return default
            raise AttributeError('missing element ' + tag)
        if len(nodes) > 1:
            raise AttributeError('found multiple %s elements, expecting only one' % tag)

        node = nodes[0]
        if isValue:
            value = ' '.join([c.nodeValue
                for c in node.childNodes
                if c.nodeType==3])
            return klass(value)
        else:
            return klass(node)

    return property(fget=accessor, doc=docstring(doc, klass))

class XmlList(object):
    def __init__(self, nodes, klass):
        self.nodes = nodes
        self.klass = klass
    def __getitem__(self, index):
        return self.klass(self.nodes[index])
    def __len__(self):
        return len(self.nodes)

def XmlListProperty(tag, doc=_marker, klass=_marker):
    def accessor(self):
        nodelist = self._node.getElementsByTagName(tag)
        return XmlList(nodelist, klass)

    return property(fget=accessor, doc=docstring(doc, klass, 'list of '))

def XmlAttributeProperty(attr, doc, typecast=unicode, default=_marker):
    '''A property which returns an attribute'''
    def accessor(self):
        node = self._node.getAttribute(attr)
        if node=='' and default is not _marker:
            return default
        return typecast(node)

    return property(fget=accessor, doc=docstring(doc, typecast))

def XmlDictProperty(tag, keyattr, klass):
    """Convert a collection of nodes into a dict"""
    def accessor(self):
        nodes = self._node.getElementsByTagName(tag)
        d = {}
        for n in nodes:
            key = n.getAttribute(keyattr)
            d[key] = klass(n)
        return d

    return property(fget=accessor, doc=docstring(klass.__doc__, klass, 'dict (keyed on %s attribute) of '%keyattr))

class XmlWrapper(object):
    __allow_access_to_unprotected_subobjects__ = 1

    def __init__(self, node):
        self._node = node

    def __str__(self):
        return ' '.join([c.nodeValue
            for c in self._node.childNodes
            if c.nodeType==3])
