##############################################################################
#
# Copyright (c) 2003-2005 Oxfam. All rights reserved.
#
# This software is distributed under the terms of the GNU Lesser
# General Public License (LGPL) License. See LICENSE.txt for license
# text.
#
##############################################################################
"""Tests for manipulating query strings, both Google and Plone

$Id$
"""

import os, sys, time
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

import Acquisition
from Testing import ZopeTestCase
from Products.CMFPlone.tests import PloneTestCase
from Products.CMFPlone.tests.PloneTestCase import portal_name, portal_owner
from AccessControl.SecurityManagement import newSecurityManager, noSecurityManager
from Products.GoogleSATool.tests import z27transaction as transaction

import Products.GoogleSATool
from Products.GoogleSATool import GoogleResults, googlesa_globals
from Products.GoogleSATool.config import GoogleError
from Globals import package_home
from DateTime import DateTime

def installGoogleSA(quiet=0):
    _start = time.time()

    ZopeTestCase.installProduct('GoogleSATool')

    # Install GoogleSATool into the test site.
    app = ZopeTestCase.app()
    user = app.acl_users.getUserById(portal_owner).__of__(app.acl_users)
    newSecurityManager(None, user)

    portal = app[portal_name]
    quickinstaller = portal.portal_quickinstaller
    quickinstaller.installProduct('GoogleSATool')

    # Log out
    noSecurityManager()
    transaction.commit()
    ZopeTestCase.close(app)

installGoogleSA()

class Field:
    def __init__(self, name, default, required):
        self.name = name
        self.default = default
        self.required = required

# Changes from default fields for testing
TEST_FIELDS = [
    Field('client', 'string:GCLIENT', True),
    Field('site', 'request/site|string:GSITE', True),
    ]

BAD_TEST_FIELDS = [
    Field('client', '', True),
    Field('site', '', True),
    ]

class TestGoogleSATool(PloneTestCase.PloneTestCase):
    """Test the tools utility functions"""
    def afterSetUp(self):
        self.tool = self.portal.portal_googlesa
        self.tool._collectionRegex = 'GSITE|other'
        REQUEST = self.portal.REQUEST
        REQUEST.form['SearchableText'] = 'water'
        REQUEST.form['submit.x'] = '0'
        REQUEST.form['submit.y'] = '0'

    def test_mq1(self):
        url = self.portal.REQUEST.get('URL0')
        query = self.tool.make_query()
        expected = url + '?SearchableText=water'
        self.assertEqual(expected, query)

    def assertQueryMatch(self, actual, expected):
        '''Compare two query strings ignoring order.'''
        aUrl, aExpect = '', ''
        
        if isinstance(actual, basestring):
            if '?' in actual:
                aUrl, actual = actual.split('?', 1)
            actual = actual.split('&')
        if isinstance(expected, basestring):
            if '?' in expected:
                aExpect, expected = expected.split('?', 1)
            expected = expected.split('&')

        self.assertEqual(aUrl, aExpect)

        for a in actual:
            try:
                pos = expected.index(a)
            except ValueError:
                raise AssertionError('Query string contained unexpected element: '+a)
            del expected[pos]
        
        if expected:
            raise AssertionError('Query string was missing:' + '&'.join(expected))

    def test_mq2(self):
        url = self.portal.REQUEST.get('URL0')
        query = self.tool.make_query(start=5, num=7)
        self.assertQueryMatch(query.split('?',1)[-1],
            'SearchableText=water&start:int=5&num:int=7')

    def test_mq_nulling(self):
        url = self.portal.REQUEST.get('URL0')
        query = self.tool.make_query(SearchableText=None, start=5, num=7)
        self.assertQueryMatch(query.split('?',1)[-1], 'start:int=5&num:int=7')
        # Try nulling everything
        query = self.tool.make_query(SearchableText=None)
        assert '?' not in query

    # Tests for building Google queries
    def test_gq(self):
        context = self.portal
        self.tool.manage_setGooglePrefs(googleURL='http://mygoogle', fields=TEST_FIELDS)

        query = self.tool.google_query(context)
        self.assertQueryMatch(query, 'http://mygoogle?q=water&site=GSITE&client=GCLIENT&output=xml&oe=utf-8')

        # Check we can override form fields here
        query = self.tool.google_query(context, q='sand', num=50)
        self.assertQueryMatch(query, 'http://mygoogle?q=sand&num=50&site=GSITE&client=GCLIENT&output=xml&oe=utf-8')

        # Cannot pass arbitrary keyword args
        self.assertRaises(GoogleError, self.tool.google_query, context, x=55)

    def test_gq_required(self):
        context = self.portal
        self.tool.manage_setGooglePrefs(googleURL='http://mygoogle', fields=BAD_TEST_FIELDS)
        self.assertRaises(GoogleError, self.tool.google_query, context)

    def test_validation(self):
        context = self.portal
        self.tool.manage_setGooglePrefs(googleURL='http://mygoogle', fields=TEST_FIELDS)
        self.portal.REQUEST.form['site'] = 'other'
        query = self.tool.google_query(context)
        self.assertQueryMatch(query, 'http://mygoogle?q=water&site=other&client=GCLIENT&output=xml&oe=utf-8')

    def test_validation2(self):
        context = self.portal
        self.tool.manage_setGooglePrefs(googleURL='http://mygoogle', fields=TEST_FIELDS)
        self.portal.REQUEST.form['site'] = 'something'
        self.assertRaises(GoogleError, self.tool.google_query, context)

    def test_validation3(self):
        context = self.portal
        self.tool.manage_setGooglePrefs(googleURL='http://mygoogle', fields=TEST_FIELDS)
        self.portal.REQUEST.form['site'] = 'otherwise'
        self.assertRaises(GoogleError, self.tool.google_query, context)
       
    def test_listarg2(self):
        context = self.portal
        self.tool.manage_setGooglePrefs(googleURL='http://mygoogle', fields=TEST_FIELDS)
        self.portal.REQUEST.form['site'] = ['other','GSITE']
        query = self.tool.google_query(context)
        self.assertQueryMatch(query, 'http://mygoogle?q=water&site=other%7CGSITE&client=GCLIENT&output=xml&oe=utf-8')
       
    def test_listarg3(self):
        context = self.portal
        self.tool.manage_setGooglePrefs(googleURL='http://mygoogle', fields=TEST_FIELDS)
        self.portal.REQUEST.form['partialfields'] = ['DC.subject:fred','DC.format:text/html']
        query = self.tool.google_query(context)
        self.assertQueryMatch(query,
            'http://mygoogle?q=water&site=GSITE&client=GCLIENT&output=xml&partialfields=DC%252Esubject%3Afred.DC%252Eformat%3Atext%252Fhtml&oe=utf-8')


if __name__ == '__main__':
    framework()
else:
    # While framework.py provides its own test_suite()
    # method the testrunner utility does not.
    from unittest import TestSuite, makeSuite
    def test_suite():
        suite = TestSuite()
        suite.addTest(makeSuite(TestGoogleSATool))
        return suite
